# Generated by Django 3.2.19 on 2024-05-24 12:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0060_workspace_chain'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='supersedes',
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='superseded',
                to='db.workrequest',
            ),
        ),
    ]
