# Generated by Django 4.2.15 on 2024-12-13 16:34

from django.db import migrations

import debusine.db.models.auth


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0113_workrequest_output_data_json'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='user',
            managers=[
                ('objects', debusine.db.models.auth.UserManager()),
            ],
        ),
    ]
