# Generated by Django 4.2.19 on 2025-04-11 09:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0141_add_clientenroll'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='workflows_need_update',
            field=models.BooleanField(default=False),
        ),
        migrations.AddIndex(
            model_name='workrequest',
            index=models.Index(
                models.F('workflows_need_update'),
                name='db_workrequest_wf_update_idx',
            ),
        ),
    ]
