# Generated by Django 3.2.19 on 2024-03-27 15:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0027_workrequest_task_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='dependencies',
            field=models.ManyToManyField(
                related_name='reverse_dependencies', to='db.WorkRequest'
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='event_reactions',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='parent',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to='db.workrequest',
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='unblock_strategy',
            field=models.CharField(
                choices=[
                    ('deps', 'Dependencies have completed'),
                    ('manual', 'Manually unblocked'),
                ],
                default='deps',
                editable=False,
                max_length=6,
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='workflow_data',
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='status',
            field=models.CharField(
                choices=[
                    ('pending', 'Pending'),
                    ('running', 'Running'),
                    ('completed', 'Completed'),
                    ('aborted', 'Aborted'),
                    ('blocked', 'Blocked'),
                ],
                default='pending',
                editable=False,
                max_length=9,
            ),
        ),
    ]
