# Generated by Django 4.2.15 on 2024-12-16 10:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0108_fill_scope_label'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scope',
            name='label',
            field=models.CharField(
                help_text='User-visible name for the scope',
                max_length=255,
                unique=True,
            ),
        ),
    ]
