# Generated by Django 4.2.18 on 2025-02-18 15:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0132_filestore_size_columns_biginteger'),
    ]

    operations = [
        migrations.AddField(
            model_name='filestore',
            name='provider_account',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='db.asset',
            ),
        ),
        migrations.AlterField(
            model_name='filestore',
            name='backend',
            field=models.CharField(
                choices=[
                    ('Local', 'Local'),
                    ('ExternalDebianSuite', 'ExternalDebianSuite'),
                    ('Memory', 'Memory'),
                    ('S3', 'S3'),
                ],
                max_length=255,
            ),
        ),
    ]
