# Generated by Django 3.2.19 on 2024-06-17 11:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0066_signing_worker_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='filestore',
            name='backend',
            field=models.CharField(
                choices=[
                    ('Local', 'Local'),
                    ('ExternalDebianSuite', 'ExternalDebianSuite'),
                    ('Memory', 'Memory'),
                ],
                max_length=255,
            ),
        ),
    ]
