# Generated by Django 4.2.20 on 2025-03-25 11:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0140_groupauditlog'),
    ]

    operations = [
        migrations.CreateModel(
            name='ClientEnroll',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('nonce', models.CharField(max_length=64, unique=True)),
                ('payload', models.JSONField(default=dict)),
            ],
        ),
    ]
