# Generated by Django 4.2.15 on 2025-01-28 12:00

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0120_scope_file_stores_data'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='workspace',
            name='default_file_store',
        ),
        migrations.RemoveField(
            model_name='workspace',
            name='other_file_stores',
        ),
    ]
