# Generated by Django 3.2.12 on 2023-05-19 08:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0004_add_field_user_to_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='created_by',
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                to='db.token',
            ),
            preserve_default=False,
        ),
    ]
