# Generated by Django 3.2.19 on 2024-07-15 12:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0067_alter_filestore_backend'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collection',
            name='workflow',
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='internal_collection',
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='workflow',
                to='db.collection',
            ),
        ),
    ]
