# Generated by Django 3.2.12 on 2023-06-12 14:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0006_alter_workrequest_created_by'),
    ]

    operations = [
        migrations.CreateModel(
            name='NotificationChannel',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=20, unique=True)),
                (
                    'method',
                    models.CharField(
                        choices=[('email', 'Email')], max_length=10
                    ),
                ),
                ('data', models.JSONField(blank=True, default=dict)),
            ],
        ),
    ]
