# Generated by Django 3.2.19 on 2024-05-22 09:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0058_workrequest_dynamic_task_data'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='collection',
            name='db_collection_unique_name_category',
        ),
        migrations.AddConstraint(
            model_name='collection',
            constraint=models.UniqueConstraint(
                fields=('name', 'category', 'workspace'),
                name='db_collection_unique_name_category_workspace',
            ),
        ),
    ]
