# Generated by Django 3.2.19 on 2024-04-22 10:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("db", "0036_workrequest_fix_task_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="workrequest",
            name="internal_collection",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="db.collection",
            ),
        ),
    ]
