# Generated by Django 3.2.19 on 2024-05-17 13:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0057_workrequest_expiration_delay'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='dynamic_task_data',
            field=models.JSONField(blank=True, null=True),
        ),
    ]
