# Generated by Django 4.2.15 on 2024-12-13 11:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0105_wait_needs_input'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='workflow_template',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to='db.workflowtemplate',
            ),
        ),
    ]
