# Generated by Django 3.2.19 on 2024-07-08 14:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0071_collectionitem_parent_category_final'),
    ]

    operations = [
        migrations.AddField(
            model_name='worker',
            name='last_seen_at',
            field=models.DateTimeField(
                blank=True,
                help_text='Last interaction between the worker and the server',
                null=True,
            ),
        ),
    ]
