# Generated by Django 3.2.19 on 2024-08-30 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0073_token_last_seen_at'),
    ]

    operations = [
        migrations.AddField(
            model_name='fileinartifact',
            name='complete',
            field=models.BooleanField(blank=True, null=True),
        ),
    ]
