# Generated by Django 3.2.19 on 2024-02-22 17:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0014_artifact_underscore_names'),
    ]

    operations = [
        # This removes null=True, which would normally require doing
        # something to populate existing rows; but it appears that there has
        # never been any way to create such rows in production code, only in
        # test code.
        migrations.AlterField(
            model_name='fileinartifact',
            name='file',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to='db.file'
            ),
        ),
    ]
