# Generated by Django 4.2.20 on 2025-03-17 18:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0138_worker_activation_token'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='groupmembership',
            constraint=models.UniqueConstraint(
                fields=('group', 'user'),
                name='db_groupmembership_unique_group_user',
            ),
        ),
    ]
