# Generated by Django 3.2.19 on 2024-09-19 08:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0079_sbuild_binnmu_build_components'),
    ]

    operations = [
        migrations.CreateModel(
            name='Scope',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=255, unique=True)),
            ],
        ),
    ]
