{% extends base_template %}
{% block content %}
    <h1>{{ title }}</h1>
    <p>You have been logged out.</p>
    <p>
        <a href="{% url "homepage:homepage" %}">Homepage</a>
    </p>
{% endblock %}
