# Generated by Django 3.2.19 on 2024-04-08 11:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0032_workrequest_manage_priorities_permission'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='collectionitem',
            name='db_collectionitem_unique_active_name',
        ),
        migrations.AddConstraint(
            model_name='collectionitem',
            constraint=models.UniqueConstraint(
                condition=models.Q(('removed_at__isnull', True)),
                fields=('name', 'parent_collection'),
                name='db_collectionitem_unique_active_name',
            ),
        ),
    ]
